
const uint ALTERED_NADEO = 24231;

bool showInterface = false;

[Setting name="Only current season" category="General"]
bool onlyCurrentSeason = true;

[Setting name="Target medal" category="General"]
Medals targetMedal = Medals::Author;

[Setting name="Automatically respin missing maps" category="General"]
bool autoRespinFails = false;

[Setting name="Hide when UI hidden" category="General"]
bool hideWithUI = false;

[Setting name="Spin volume" category="General" min=0.f max=1.f]
float spinVolume = 0.1f;

[Setting name="Selected club" category="Advanced"]
uint selected_club = ALTERED_NADEO;
[Setting name="Ignore map name numbers in other clubs" description="Use the map's position in a campaign as its number. Missing/unvalidated maps will cause the wrong map to be loaded, but this allows you to use your own club campaigns with different numbering" category="Advanced"]
bool useNameNumbers = true;

[Setting name="debug print" category="Advanced"]
bool debugPrint = false;

UI::Font@ largeFont = UI::LoadFont("DroidSans.ttf", 26);
UI::Font@ mediumFont = UI::LoadFont("DroidSans.ttf", 20);
UI::Font@ digitalFont = UI::LoadFont("digital-7.mono.ttf", 20);

Audio::Sample@ spinSound = Audio::LoadSample('spin.wav');


string selected_club_name = '';


void PrintActivityArray() {
    for (uint i = 0; i < activities.Length; i++) {
        print("" + activities[i].ToString());
    }
}


void RenderMenu() {
    if (UI::MenuItem("\\$29e" + Icons::Dodecahedron + "\\$z " + Meta::ExecutingPlugin().Name, "", showInterface)) {
        showInterface = !showInterface;
	}
}

vec4 runColor = vec4(0.f, 0.9f, 0.f, 1.f), settingsColor = vec4(0.9f, 0.f, 0.9f, 1.f), stateColor = vec4(0.9f, 0.f, 0.f, 1.f);
uint8 selectedTab = 0;

void Render() {
    if (!showInterface) {return;}
    if (hideWithUI && !UI::IsGameUIVisible() && GetApp().CurrentPlayground !is null) {return;}
    
    UI::SetNextWindowSize(0, 0, UI::Cond::Always);
    int flags;
    if (UI::Begin(Meta::ExecutingPlugin().Name, showInterface, UI::WindowFlags::NoCollapse)) {
        UI::BeginTabBar(Meta::ExecutingPlugin().Name, UI::TabBarFlags::NoCloseWithMiddleMouseButton);
        flags = UI::TabItemFlags::None;
        if (selectedTab == 1) {flags |= UI::TabItemFlags::SetSelected; selectedTab = 0;}
        if (UI::BeginTabItem('Info', flags)) {
            renderInfo();
            UI::EndTabItem();
        }
        UI::PushStyleColor(UI::Col::Tab, runColor * vec4(0.5, 0.5, 0.5, 0.75));
        UI::PushStyleColor(UI::Col::TabHovered, runColor * vec4(1.2, 1.2, 1.2, 0.85));
        UI::PushStyleColor(UI::Col::TabActive, runColor);
        flags = UI::TabItemFlags::None;
        if (selectedTab == 2) {flags |= UI::TabItemFlags::SetSelected; selectedTab = 0;}
        if (UI::BeginTabItem('Run', flags)) {
            Run::render();
            UI::EndTabItem();
        }
        UI::PopStyleColor(3);
        UI::PushStyleColor(UI::Col::Tab, settingsColor * vec4(0.5, 0.5, 0.5, 0.75));
        UI::PushStyleColor(UI::Col::TabHovered, settingsColor * vec4(1.2, 1.2, 1.2, 0.85));
        UI::PushStyleColor(UI::Col::TabActive, settingsColor);
        flags = UI::TabItemFlags::None;
        if (selectedTab == 3) {flags |= UI::TabItemFlags::SetSelected; selectedTab = 0;}
        if (UI::BeginTabItem('Settings', flags)) {
            if (process == Process::BadClubId) {
                UI::Text('Selected Club: unknown club');
            } else if (process != Process::UnknownClub) {
                UI::Text('Selected Club: ' + selected_club_name);
            } else {
                UI::Text('Selected Club: getting club name...');
            }
            Run::renderSettings();
            UI::EndTabItem();
        }
        UI::PopStyleColor(3);
        UI::PushStyleColor(UI::Col::Tab, stateColor * vec4(0.5, 0.5, 0.5, 0.75));
        UI::PushStyleColor(UI::Col::TabHovered, stateColor * vec4(1.2, 1.2, 1.2, 0.85));
        UI::PushStyleColor(UI::Col::TabActive, stateColor);
        flags = UI::TabItemFlags::None;
        if (selectedTab == 4) {flags |= UI::TabItemFlags::SetSelected; selectedTab = 0;}
        if (UI::BeginTabItem('State', flags)) {
            Run::renderStateSettings();
            UI::EndTabItem();
        }
        UI::PopStyleColor(3);
        UI::EndTabBar();

    }
    UI::End();
}

string formatTime(uint64 time) {
    string t = Time::Format(time);
    if (t.Length == 8) {
        t = '0'+t;
    }
    return t;
}

void renderInfo() {
    UI::Text('Welcome to Random Altered Campaign Challenge.');
    UI::TextWrapped('This challenge is to beat [' + Run::getMedalName(targetMedal) + '\\$z] on 25 maps, each picked from a random Altered Campaign.');
    UI::TextWrapped('Similar to RMC/RMS, it is possible to reroll (skip).\n' + 
        'If you get [' + Run::getRerollMedalName(targetMedal) + '\\$z] or ' + Run::getMedalName(Medals::WorldRecord) + '\\$z you can reroll for free.');
    UI::TextWrapped('You can also reroll any map at any time, however this adds one\\$f00' + Icons::Lock + '\\$z. You cannot move on to the next number with any\\$f00' + Icons::Lock + '\\$zand each [' + Run::getMedalName(targetMedal) + '\\$z] will instead remove 1\\$f00' + Icons::Lock + '\\$z.');
    UI::TextWrapped('If a map is hacked/unvalidated, and the map has no records, then you may broken reroll it. Maps slowed by patches may also be free rerolled.');
    UI::TextWrapped('However, there is no map length or type restrictions so you must play long, stunt, platform, dragonyeet, earthquake, etc. maps. \n(Or\\$f00' + Icons::Lock + '\\$zreroll)');
    UI::TextWrapped('Watching World Record is not allowed until you have been on the map for 15 minutes');
    UI::TextWrapped('Unclaimed Checkpoint Indicator plugin is allowed (and recommended for Checkpointfull alt campaign)');
    UI::Separator();
    UI::TextWrapped('Maps are selected by spinning a \'wheel\' inside the plugin.');
    UI::TextWrapped('You will need to press a button to spin for each map, and the plugin lets the spinner finish before loading into the map. However, there [will be] settings to reduce this animation');

}