
class Map {
    string uid;
    string name;
    string simpleName;
    string type;
    string url;
    Activity@ source;
    uint number;
    string expectedName;

    Map(const string &in name, const string &in uid, const string &in type, const string &in url, Activity@ source) {
        this.name = name;
        this.uid = uid;
        this.simpleName = Text::StripOpenplanetFormatCodes(name);
        this.type = type;
        this.url = url;
        @this.source = source;
        this.number = getNumber();
        this.expectedName = source.getSeason() + ' - ' + Text::Format('%02d', this.number) + ' ' + source.specialAlteration + (source.specialAlteration == '' ? '' : ' ') + getAlterationName(source.alteration, source.originalname);
        // if (debugPrint) {print('cached map: ' + this.expectedName);}
    }

    void play() {
        if (!Permissions::PlayLocalMap()) {
            return;
        }
        ReturnToMenu();
        string mode = this.type.ToLower();
        if (mode.EndsWith("race")) {
            mode = "TrackMania\\TM_PlayMap_Local";
        } else if (mode.EndsWith("stunt")) {
            mode = "Trackmania\\TM_StuntSolo_Local";
        } else if (mode.EndsWith("platform")) {
            mode = "Trackmania\\TM_Platform_Local";
        } else if (mode.EndsWith("royal")) {
            mode = "Trackmania\\TM_RoyalTimeAttack_Local";
        } else {
            print('unknown mode: ' + mode);
            mode = "";
        }
        print('loaded map: ' + this.name + ' with mode: ' + mode);
        
        CGameManiaPlanet@ app = cast<CGameManiaPlanet>(GetApp());
        uint64 loadtime = Time::Now;
        while (!app.ManiaTitleControlScriptAPI.IsReady) {
            if (Time::Now - loadtime > 10000) {
                UI::ShowNotification(Meta::ExecutingPlugin().Name, "Timed out on loading map: " + this.expectedName, vec4(0.5,0,0,0.5));
                return;
            } else {
                yield();
            }
        }
        app.ManiaTitleControlScriptAPI.PlayMap(url, mode, "");
    }

    uint getNumber() {
        // discovery campaigns don't use map numbers
        if ((this.source.season == 0 && this.source.year == 2024)) {
            return getSnowDiscoveryNumber(this.simpleName.ToLower());
        }
        if ((this.source.season == 4 && this.source.year == 2024)) {
            return getRallyDiscoveryNumber(this.simpleName.ToLower());
        }
        if ((this.source.season == 8 && this.source.year == 2024)) {
            return getDesertDiscoveryNumber(this.simpleName.ToLower());
        }
        if ((this.source.season == 12 && this.source.year == 2024)) {
            return getStuntDiscoveryNumber(this.simpleName.ToLower());
        }
        if ((this.source.season == 0 && this.source.year == 2025)) {
            return getPlatformDiscoveryNumber(this.simpleName.ToLower());
        }
        if ((this.source.season == 8 && this.source.year == 2021)) {
            return getRoyalNumber(this.simpleName.ToLower());
        }
        // 
        uint number = 0;
        uint numfound = 0;
        for (uint i = 1; i <= 25; i++) {
            if (this.simpleName.Contains('- ' + Text::Format('%02d', i))) {
                number = i;
                numfound += 1;
            }
        }
        if (numfound == 1) {
            return number;
        } else if (numfound > 1) {
            number = getDuplicateMapNumber(this); // ReallyMad a08 ' - 08' campaign
            if (number == 0) {
                if (debugPrint) {print('invalid map name: ' + this.simpleName);}
            }
            return number;
        }
        // alpha campaign
        if (this.source.season == 7 && this.source.year == 2020) {
            for (uint i = 0; i <= 10; i++) {
                if (this.simpleName.Contains('T' + Text::Format('%02d', i))) {
                    number = i;
                    numfound += 1;
                }
            }
            for (uint i = 0; i <= 15; i++) {
                if (this.simpleName.Contains('S' + Text::Format('%02d', i))) {
                    number = i+10;
                    numfound += 1;
                }
            }
            if (numfound == 1) {
                return number;
            } else if (numfound > 1) {
                if (debugPrint) {print('invalid map name: ' + this.simpleName);}
                return 0;
            }
        }
        // check extras
        number = getSpecialMapNumber(this);
        if (number == 0) {
            if (debugPrint) {print('invalid map name: ' + this.simpleName);}
        }
        return number;
    }

}


void SetMapsFromUidList(Activity@ activity, array<string>@ responsemaps) {
    array<Map@>@ newmaps = {};
    // create a request to get the names of the maps
    string mapnamesurl = NadeoServices::BaseURLLive() + '/api/token/map/get-multiple?mapUidList=' + responsemaps[0];
    // nadeo has a limit of 100 maps per request
    for (uint i = 1; i < responsemaps.Length && i < 100; i++) { 
        mapnamesurl += ',' + responsemaps[i];
    }
    // request the info (for names) of the maps
    Net::HttpRequest@ req_names = NadeoServices::Get("NadeoLiveServices", mapnamesurl);
    req_names.Start();
    while(!req_names.Finished()) {yield();}
    if (req_names.ResponseCode() != 200) {
        print('invalid http response when getting activity maps map names' + req_names.ResponseCode() + ', ' + req_names.String());
    }
    Json::Value@ names_response = Json::Parse(req_names.String());
    if (!names_response.HasKey('mapList')) {
        print('invalid response when getting activity maps map names ' + req_names.String());
        return;
    }
    Json::Value@ infolist = names_response['mapList'];

    // I want to keep the order from the original request and don't know if the get-multiple is order consistent
    bool found;
    for (uint i = 0; i < responsemaps.Length; i++) {
        // order often seems consistent so check that first, but if not then check all
        if (i < infolist.Length && infolist[i]['uid'] == responsemaps[i]) {
            newmaps.InsertLast(Map(Text::OpenplanetFormatCodes(infolist[i]['name']), responsemaps[i], infolist[i]['mapType'], infolist[i]['downloadUrl'], activity));
        } else {
            found = false;
            for (uint j = 0; j < infolist.Length; j++) {
                if (j == i) {continue;} // this is pre-checked anyway (not like string comparison is expensive tho)
                if (infolist[j]['uid'] == responsemaps[i]) {
                    newmaps.InsertLast(Map(Text::OpenplanetFormatCodes(infolist[j]['name']), responsemaps[i], infolist[j]['mapType'], infolist[j]['downloadUrl'], activity));
                    found = true;
                    break;
                }
            }
            if (!found) {
                print('no info response for map: ' + responsemaps[i]);
            }
        }
    }
    activityMaps[tostring(activity.campaign)] = newmaps;
}

Map@ getMapForNumber(const string &in campaign, uint number) {
    array<Map@>@ maplist;
    activityMaps.Get(campaign, @maplist);
    if (selected_club != ALTERED_NADEO && !useNameNumbers) {
        if (maplist.Length > number) {
            return maplist[number];
        }
        return null;
    }
    for (uint i = number - 1; i < maplist.Length; i = i == 0 ? number : (i < number ? i - 1 : i + 1)) {
        if (maplist[i].number == number) {
            return maplist[i];
        }
    }
    return null;
}
