
enum Medals {
    NoMedal,
    Bronze,
    Silver,
    Gold,
    Author,
    AutoBronze,
    AutoSilver,
    AutoGold,
    AutoAuthor,
    WorldRecord

}

bool allowEditState = false;
int editCampaignId = 0;

namespace Run {
    // current map number playing
    uint mapNum = 0;
    // total maps played including rerolls and locks
    uint mapsPlayed = 0;
    // number of rerolls by taking medal (gold by default)
    uint medalRerollCount = 0;
    // number of rerolls by taking rank (wr by default)
    uint rankRerollCount = 0;
    // 'free' reroll adds an extra [target medal] required to move on
    uint extrasRequired = 0;
    // time spent on challenge
    uint64 duration = 0;
    //time spent on current map
    uint64 mapduration = 0;

    bool paused = false;
    bool brokenSkipUI = false;

    // time of last update run (for accurate timer pausing while not driving)
    uint64 lastUpdateTime;
    // time needed to reroll based on leaderboard rank (default WR)
    uint rankRerollTime = uint(-1);
    // time needed to reroll based on medal (default Gold)
    uint medalRerollTime = uint(-1);
    // time needed to achieve map
    uint objectiveTime = uint(-1);
    // last seen correct map id (for getting scores on new map)
    string savedMap = '';
    // map pb during challenge
    uint mapPB = uint(-1);
    // if current map gamemode wants higher score value to beat AT (e.g. stunt mode)
    bool highBetter;
    FinishData finishDataType;
    // map object of current map (reference for name)
    Map@ currentMap = null;

    enum Status {
        // before start run is pressed, or after a run is finished
        NotStarted,
        // pressed start and getting activities
        WaitingToStart,
        // set from ui, will roll map next update
        ToRollMap,
        // deciding on a map and verifying it is available
        RollingMap,
        // playing a map (or loading into a map)
        Playing,
        // beaten target medal (or skipped) but not automatically moving on so you can watch wr
        ManualContinue,
        // failed to load map (e.g. missing number in campaign), asks to spin again
        ManualReroll
    }

    Status status = Status::NotStarted;

    void start() {
        if (!Permissions::PlayLocalMap()) {
            return;
        }
        mapNum = 0;
        medalRerollCount = 0;
        rankRerollCount = 0;
        extrasRequired = 0;
        mapsPlayed = 0;
        targetMedal = targetMedal;
        duration = 0;
        mapduration = 0;
        paused = false;
        mapPB = uint(-1);
        mapNum = 0;
        status = Status::WaitingToStart;
        process = Process::GetActivities;
    }
    void updateMaps() {
        Wheel::fromActivities();
        if (status == Status::WaitingToStart) {
            status = Status::ManualContinue;
        }
    }

    void onNewMap(bool reopen = false) {
        if (!reopen) {
            mapPB = uint(-1);
            mapduration = 0;
        }
        highBetter = getHighBetter();
        finishDataType = getFinishDataType();
        rankRerollTime = getRankTime(); //might need startnew wrapper function?
        medalRerollTime = getRerollMedalTime();
        objectiveTime = getObjectiveTime();
    }

    void render() {
        if (status == Status::NotStarted || status == Status::WaitingToStart) {
            if (status == Status::NotStarted && (process == Process::HasActivities || process == Process::NoActivities)) {
                if (UI::ButtonColored('Start Random Altered Campaign Challenge', 0.33f)) {
                    start();
                }
            } else {
                UI::BeginDisabled();
                UI::ButtonColored('Start Random Altered Campaign Challenge', 0.33f);
                UI::EndDisabled();
            }
        } else {
            if (UI::ButtonColored('Stop Random Altered Campaign Challenge', 0.f)) {
                status = Status::NotStarted;
            }
        }
        UI::Separator();

        UI::PushFont(digitalFont);
        UI::Text(formatTime(duration));
        UI::PopFont();
        UI::Dummy(vec2(0.f, 4.f * UI::GetScale()));
        UI::Text(Icons::Map + ' ' + formatTime(mapduration));
        UI::Separator();
        UI::PushFont(largeFont);
        if (mapNum == 25) {
            UI::Text('Challenge completed!');
        } else if (extrasRequired == 0) {
            UI::Text(getMedalIcon() + '\\$z Progress: ' + mapNum);
        } else {
            UI::Text(getMedalIcon() + '\\$z Progress: ' + mapNum + ' - \\$f00' + Icons::Lock + '\\$z: ' + extrasRequired);
        }
        // color of reroll medal (which is not targetMedal)
        UI::PopFont();
        UI::PushFont(mediumFont);
        UI::Text(getRerollMedalIcon() + '\\$fff: ' + medalRerollCount + '  WR: ' + rankRerollCount);
        UI::PopFont();
        UI::SameLine();
        UI::SetCursorPos(UI::GetCursorPos() + vec2(0.f, 2.f * UI::GetScale()));
        UI::Text('Maps Played: ' + mapsPlayed);
        UI::Separator();

        bool renderwheel = true;
        if (status == Status::NotStarted) {
            renderwheel = false;
        } else if (status == Status::WaitingToStart || process == Process::GetActivities) {
            renderwheel = false;
            UI::Text('Loading maps from club...');
        } else if (status == Status::ManualContinue) {
            UI::SetCursorPos(UI::GetCursorPos() + vec2(32 * UI::GetScale(), 0));
            float coloradjust = Math::Abs(500.f - float(Time::Now % 1000)) / 500;
            if (UI::ButtonColored('Roll new map!', 0.15f - coloradjust * 0.1f, 1.f - (coloradjust * 0.75))) {
                status = Status::ToRollMap;
            }
        } else if (status == Status::ManualReroll && autoRespinFails) {
            status = Status::ToRollMap;
        } else if (status == Status::ManualReroll && Wheel::targetTime <= Time::Now) {
            UI::SetCursorPos(UI::GetCursorPos() + vec2(32 * UI::GetScale(), 0));
            float coloradjust = Math::Abs(500.f - float(Time::Now % 1000)) / 500;
            if (UI::ButtonColored('Reroll new map!', 0.15f - coloradjust * 0.1f, 1.f - (coloradjust * 0.75))) {
                status = Status::ToRollMap;
            }
            UI::SameLine();
            UI::Text('No map found for this number');
        } else if (status == Status::ToRollMap || status == Status::RollingMap || status == Status::ManualReroll) {
            UI::SetCursorPos(UI::GetCursorPos() + vec2(32 * UI::GetScale(), 0));
            UI::BeginDisabled();
            UI::ButtonColored('Rolling...', 0.15);
            UI::EndDisabled();
            // UI::Text('Rolling map...');
        } else if (status == Status::Playing) {
            CGameCtnApp@ app = GetApp();
            CGameCtnChallenge@ rootMap = cast<CGameCtnChallenge>(GetApp().RootMap);
            CGamePlayground@ playground = cast<CGamePlayground>(app.CurrentPlayground);
            if (playground is null || rootMap is null || currentMap is null || rootMap.IdName != currentMap.uid) {
                UI::SetCursorPos(UI::GetCursorPos() + vec2(32 * UI::GetScale(), 0));
                UI::BeginDisabled();
                UI::ButtonColored('Switching map...', 0.15);
                UI::EndDisabled();
                if (rootMap !is null && currentMap !is null && rootMap.IdName != currentMap.uid) {
                    UI::SameLine();
                    UI::Text('Current map doesn\'t match');
                }
            } else {
                if (currentMap !is null) {
                    UI::Text(currentMap.expectedName);
                    UI::SameLine();
                    UI::TextDisabled('by ' + rootMap.AuthorNickName);
                    string exever;
                    try {
                        exever = Dev::GetOffsetString(rootMap, 0x78).Split('_', 2)[0].SubStr(5);
                    } catch {
                        exever = 'Unknown';
                    }
                    if (exever == 'Unknown') {
                    } else if (exever < "2022-05-19") {
                        UI::TextDisabled('Game version: ' + exever);
                        UI::SameLine();
                        UI::Text('(before ice patches)');
                    } else if (exever < "2023-04-289") {
                        UI::TextDisabled('Game version: ' + exever);
                        UI::SameLine();
                        UI::Text('(between ice patches)');
                    } else {
                        UI::TextDisabled('Game version: ' + exever);
                        UI::SameLine();
                        UI::TextDisabled('(after ice patches)');
                    }
                }
                UI::Separator();
                renderwheel = false;
                if (paused) {
                    if (UI::Button(Icons::HourglassO + Icons::Play, vec2(0,0))) {
                        paused = false;
                    }
                } else {
                    uint8 sandvalue = ((mapduration/690) % 3);
                    if (UI::Button((sandvalue == 0 ? Icons::HourglassStart : sandvalue == 1 ? Icons::HourglassHalf : Icons::HourglassEnd) + Icons::Pause, vec2(0,0))) {
                        paused = true;
                    }
                }
                if (medalRerollTime == uint(-1) || rankRerollTime == uint(-1) || (highBetter && rankRerollTime >= medalRerollTime || !highBetter && rankRerollTime <= medalRerollTime)) {
                    UI::SameLine();
                    if (medalRerollTime == uint(-1) || ((mapPB != uint(-1) || targetMedal == Medals::NoMedal) && (highBetter && mapPB >= medalRerollTime || !highBetter && mapPB <= medalRerollTime))) {
                        if (UI::Button(getRerollMedalName() + '\\$z reroll')) {
                            medalRerollCount += 1;
                            mapsPlayed += 1;
                            status = Status::ManualContinue;
                            Wheel::reset();
                        }
                    } else {
                        UI::BeginDisabled();
                        UI::Button(getRerollMedalName() + '\\$fff reroll');
                        UI::EndDisabled();
                    }
                }
                if (medalRerollTime == uint(-1) || rankRerollTime == uint(-1) || (highBetter && rankRerollTime < medalRerollTime || !highBetter && rankRerollTime > medalRerollTime)) {
                    UI::SameLine();
                    if (rankRerollTime == uint(-1) || (mapPB != uint(-1) && (highBetter && mapPB >= rankRerollTime || !highBetter && mapPB <= rankRerollTime))) {
                        if (UI::Button(getMedalName(Medals::WorldRecord) + '\\$fff reroll')) {
                            rankRerollCount += 1;
                            mapsPlayed += 1;
                            status = Status::ManualContinue;
                            Wheel::reset();
                        }
                    } else {
                        UI::BeginDisabled();
                        UI::Button(getMedalName(Medals::WorldRecord) + '\\$fff reroll');
                        UI::EndDisabled();
                    }
                }
                // UI::SameLine();
                if (playground.GameTerminals.Length > 0) {
                    if (UI::Button('+1\\$f00' + Icons::Lock + '\\$zReroll')) {
                        extrasRequired += 1;
                        mapsPlayed += 1;
                        status = Status::ManualContinue;
                        Wheel::reset();
                    }
                } else {
                    UI::BeginDisabled();
                    UI::Button('+1\\$f00' + Icons::Lock + '\\$zReroll');
                    UI::EndDisabled();
                }
            }
        }
        if (renderwheel) {
            Wheel::render();
        } else if (status == Status::Playing && GetApp().CurrentPlayground is null) {
                // UI::SetCursorPos(UI::GetCursorPos() + vec2(32 * UI::GetScale(), 0));
                UI::BeginDisabled();
                UI::Text('Switching map...');
                UI::EndDisabled();
        }
        if (status == Status::Playing) {
            if (UI::Button('Reopen map')) {
                openCurrentMap();
            }
            UI::SameLine();
            if (UI::ButtonColored(Icons::PlayCircle + 'Reroll broken map', 0.155f)) {
                brokenSkipUI = true;
                UI::OpenPopup('Broken reroll');
            } 
            
            if (brokenSkipUI = UI::BeginPopupModal('Broken reroll')) {
                UI::Text('Broken map rerolls are only for impossible or broken maps.\nThe run time will be reverted to when you started this map.\nNo grip wasn\'t changed like ice.\nAre you sure you want to reroll?');
                if (UI::Button('Cancel')) {
                    brokenSkipUI = false;
                    UI::CloseCurrentPopup();
                }
                UI::SameLine();
                if (UI::ButtonColored(Icons::PlayCircle + 'Reroll broken map', 0.155f)) {
                    status = Status::ManualContinue;
                    duration -= mapduration;
                    mapsPlayed += 1;
                    Wheel::reset();
                    brokenSkipUI = false;
                    UI::CloseCurrentPopup();
                }
			    UI::EndPopup();
            }
        }
        // CGamePlayground@ playground = cast<CGamePlayground>(GetApp().CurrentPlayground);
        // if (playground !is null && playground.GameTerminals.Length > 0) {
        //     UI::Text(tostring(playground.GameTerminals[0].UISequence_Current));
        //     UI::SameLine();
        //     UI::Text(tostring(checkFinished(FinishData::Time)));
        //     UI::SameLine();
        //     UI::Text(tostring(getRerollMedalTime()));
        //     UI::SameLine();
        //     UI::Text(tostring(mapPB));
        // }
    }

    void renderSettings() {
        if ((process == Process::HasActivities || process == Process::NoActivities) && (status == Status::ManualContinue || status == Status::ManualReroll || status == Status::Playing)) {
            if (UI::Button('Refresh club activities')) {
                process == Process::GetActivities;
            }
        } else {
            UI::BeginDisabled();
            UI::Button('Refresh club activities');
            UI::EndDisabled();
        }
        onlyCurrentSeason = UI::Checkbox('##current', onlyCurrentSeason);
        UI::SameLine();
        UI::Text('Current season only');

        UI::SetNextItemWidth(100);
        Medals newtarget = targetMedal;
        if (UI::BeginCombo('##medal', getMedalName(targetMedal), UI::ComboFlags::None)) {
            if (UI::Selectable(getMedalName(Medals::WorldRecord), targetMedal == Medals::WorldRecord, UI::SelectableFlags::None)) {
                newtarget = Medals::WorldRecord;
            } else 
            if (UI::Selectable(getMedalName(Medals::Author), targetMedal == Medals::Author, UI::SelectableFlags::None)) {
                newtarget = Medals::Author;
            } else if (UI::Selectable(getMedalName(Medals::Gold), targetMedal == Medals::Gold, UI::SelectableFlags::None)) {
                newtarget = Medals::Gold;
            } else if (UI::Selectable(getMedalName(Medals::Silver), targetMedal == Medals::Silver, UI::SelectableFlags::None)) {
                newtarget = Medals::Silver;
            } else if (UI::Selectable(getMedalName(Medals::Bronze), targetMedal == Medals::Bronze, UI::SelectableFlags::None)) {
                newtarget = Medals::Bronze;
            } else if (UI::Selectable(getMedalName(Medals::NoMedal), targetMedal == Medals::NoMedal, UI::SelectableFlags::None)) {
                newtarget = Medals::NoMedal;
            } else if (UI::Selectable(getMedalName(Medals::AutoAuthor), targetMedal == Medals::AutoAuthor, UI::SelectableFlags::None)) {
                newtarget = Medals::AutoAuthor;
            } else if (UI::Selectable(getMedalName(Medals::AutoGold), targetMedal == Medals::AutoGold, UI::SelectableFlags::None)) {
                newtarget = Medals::AutoGold;
            } else if (UI::Selectable(getMedalName(Medals::AutoSilver), targetMedal == Medals::AutoSilver, UI::SelectableFlags::None)) {
                newtarget = Medals::AutoSilver;
            } else if (UI::Selectable(getMedalName(Medals::AutoBronze), targetMedal == Medals::AutoBronze, UI::SelectableFlags::None)) {
                newtarget = Medals::AutoBronze;
            }
            if (newtarget != targetMedal) {
                targetMedal = newtarget;
                if (status == Status::Playing && currentMap !is null && GetApp().RootMap !is null) {
                    medalRerollTime = getRerollMedalTime();
                }
            }
            UI::EndCombo();
        }
        UI::SameLine();
        UI::Text('Target medal');

        autoRespinFails = UI::Checkbox('##respin', autoRespinFails);
        UI::SameLine();
        UI::Text('Automatically respin missing maps');

        hideWithUI = UI::Checkbox('##hide', hideWithUI);
        UI::SameLine();
        UI::Text('Hide plugin when game UI hidden');

        UI::SetNextItemWidth(100);
        spinVolume = UI::SliderFloat('##volume', spinVolume, 0.f, 1.f);
        UI::SameLine();
        UI::Text('Spin volume');

    }
    void renderStateSettings() {
        UI::Text('These settings allow you to edit the current run.\nThis is intended for dev/testing, and\nresuming a run from a previous session.');
        if (allowEditState) {
            UI::BeginDisabled();
            UI::Button('Editing state allowed');
            UI::EndDisabled();
        } else if (status == Status::NotStarted) {
            UI::BeginDisabled();
            UI::Button('No run to edit');
            UI::SameLine();
            UI::TextDisabled('click start first');
            UI::EndDisabled();
        } else {
            if (UI::Button('Allow editing state')) {
                allowEditState = true;
            }
        }
        if (!allowEditState || status == Status::NotStarted) {
            UI::BeginDisabled();
            UI::SetNextItemWidth(100);
            UI::SliderInt('##mapNum', mapNum, 0, 25);
            UI::SameLine();
            UI::Text('Maps completed');
            UI::SetNextItemWidth(100);
            UI::SliderInt('##mapsPlayed', mapsPlayed, 0, 25);
            UI::SameLine();
            UI::Text('Maps played');
            UI::SetNextItemWidth(100);
            UI::InputInt('##medalRerollCount', medalRerollCount, 0);
            UI::SameLine();
            UI::Text('Medal reroll count');
            UI::SetNextItemWidth(100);
            UI::InputInt('##rankRerollCount', rankRerollCount, 0);
            UI::SameLine();
            UI::Text('WR reroll count');
            UI::SetNextItemWidth(100);
            UI::InputInt('##extrasRequired', extrasRequired, 0);
            UI::SameLine();
            UI::Text('Extras (\\$f00' + Icons::Lock + '\\$z) required');
            UI::SetNextItemWidth(100);
            UI::InputText('##duration', tostring(duration));
            UI::SameLine();
            UI::Text('Duration (ms)');
            UI::SetNextItemWidth(100);
            UI::InputText('##mapduration', tostring(mapduration));
            UI::SameLine();
            UI::Text('Map duration (ms)');
            UI::Separator();
            UI::SetNextItemWidth(100);
            UI::InputInt('##campaign', editCampaignId, 0);
            UI::SameLine();
            UI::Button('Set campaign id');
            UI::EndDisabled();
        } else {
            UI::SetNextItemWidth(100);
            mapNum = uint(UI::SliderInt('##mapNum', mapNum, 0, 25));
            UI::SameLine();
            UI::Text('Maps completed');
            UI::SetNextItemWidth(100);
            mapsPlayed = uint(UI::InputInt('##mapsPlayed', mapsPlayed, 0));
            UI::SameLine();
            UI::Text('Maps played');
            UI::SetNextItemWidth(100);
            medalRerollCount = uint(UI::InputInt('##medalRerollCount', medalRerollCount, 0));
            UI::SameLine();
            UI::Text('Medal reroll count');
            UI::SetNextItemWidth(100);
            rankRerollCount = uint(UI::InputInt('##rankRerollCount', rankRerollCount, 0));
            UI::SameLine();
            UI::Text('WR reroll count');
            UI::SetNextItemWidth(100);
            extrasRequired = uint(UI::InputInt('##extrasRequired', extrasRequired, 0));
            UI::SameLine();
            UI::Text('Extras (\\$f00' + Icons::Lock + '\\$z) required');
            UI::SetNextItemWidth(100);
            duration = Text::ParseUInt64(UI::InputText('##duration', tostring(duration)));
            UI::SameLine();
            UI::Text('Duration (ms)');
            UI::SetNextItemWidth(100);
            mapduration = Text::ParseUInt64(UI::InputText('##mapduration', tostring(mapduration)));
            UI::SameLine();
            UI::Text('Map duration (ms)');
            UI::Separator();
            UI::SetNextItemWidth(100);
            editCampaignId = UI::InputInt('##campaign', editCampaignId, 0);
            UI::SameLine();
            if (UI::Button('Set campaign id')) {
                if (Wheel::setTarget(editCampaignId)) {
                    @currentMap = null;
                    Wheel::startSpin(Wheel::targetindex);
                    Run::status = Run::Status::RollingMap;
                    selectedTab = 2;
                    startnew(_rollNewMapAsync);
                } else {
                    UI::ShowNotification(Meta::ExecutingPlugin().Name, "Invalid campaign", vec4(0.5,0,0,0.5));
                    editCampaignId = 0;
                }
            }
        }
        if (Run::status == Status::Playing) {
            if (UI::ButtonColored(Icons::PlayCircle + 'Reroll map (but keep time)', 0.155f)) {
                status = Status::ManualContinue;
                mapsPlayed += 1;
                Wheel::reset();
                selectedTab = 2;
            }
        } else {
            UI::BeginDisabled();
            UI::ButtonColored(Icons::PlayCircle + 'Reroll map (but keep time)', 0.155f);
            UI::EndDisabled();
        }
    }

    void update() {
        if (paused || status == Status::NotStarted) {
            lastUpdateTime = Time::Now;
            return;
        }
        CGameCtnApp@ app = GetApp();
        CGameCtnChallenge@ rootMap = cast<CGameCtnChallenge>(GetApp().RootMap);
        CGamePlayground@ playground = cast<CGamePlayground>(app.CurrentPlayground);

        if (status == Status::WaitingToStart && process == Process::HasActivities) {
            start();
        }
        
        // check if map has changed
        if (playground is null || rootMap is null || currentMap is null || rootMap.IdName != currentMap.uid) {
            savedMap = '';
        } else {
            if (savedMap != rootMap.IdName) {
                onNewMap(savedMap == currentMap.uid);
                savedMap = rootMap.IdName;
            }

            // check if race end (to save pb)
            uint pb = checkFinished(finishDataType);
            if (pb != uint(-1)) {
                if (mapPB == uint(-1) || highBetter && pb > mapPB || !highBetter && pb < mapPB) {
                    mapPB = pb;
                    if (objectiveTime == uint(-1) && targetMedal == Medals::WorldRecord) {
                        objectiveTime = rankRerollTime;
                    }
                    if (objectiveTime != uint(-1) && (highBetter && mapPB >= objectiveTime || !highBetter && mapPB <= objectiveTime)) {
                        if (extrasRequired == 0) {
                            mapNum ++;
                            mapsPlayed += 1;
                            if (mapNum == 25) {
                                status == Status::NotStarted;
                            }
                        } else {
                            extrasRequired -= 1;
                        }
                        status = Status::ManualContinue;
                        Wheel::reset();
                    }
                }
            }
        }

        // add time to current map (and duration)
        uint64 time = Time::Now;
        if (status == Status::Playing && !paused && 
            rootMap !is null && playground !is null && 
            currentMap !is null && rootMap.IdName == currentMap.uid &&
            playground.GameTerminals.Length > 0 && 
            !cast<CTrackMania>(app).ManiaPlanetScriptAPI.ActiveContext_InGameMenuDisplayed) {
                SGamePlaygroundUIConfig::EUISequence uiSequence = playground.GameTerminals[0].UISequence_Current;
                if (uiSequence == CGamePlaygroundUIConfig::EUISequence::Playing || 
                    uiSequence == CGamePlaygroundUIConfig::EUISequence::CustomMTClip || 
                    uiSequence == CGamePlaygroundUIConfig::EUISequence::CustomMTClip_WithUIInteraction) {
                        duration += (time - lastUpdateTime);
                        mapduration += (time - lastUpdateTime);
                }
        }
        lastUpdateTime = time;
    }

    void rollNewMap() {
        @currentMap = null;
        Wheel::startSpin();
        Run::status = Run::Status::RollingMap;
        startnew(_rollNewMapAsync);
    }
    void _rollNewMapAsync() {
        Activity@ target = Wheel::getTarget();
        string targetid = tostring(target.campaign);
        print('rolling wheel to: ' + target.ToString() + ' (' + targetid + ')');
        target.getMaps(); // if cached already, does nothing
        if (!activityMaps.Exists(targetid)) {
            status = Status::ManualReroll;
            UI::ShowNotification(Meta::ExecutingPlugin().Name, "Failed to get campaign: " + target.ToString(), vec4(0.5,0,0,0.5));
            return;
        }
        // check if map number is valid / get index
        @currentMap = getMapForNumber(targetid, mapNum + 1);
        if (currentMap is null) {
            if (debugPrint) {print("no map for number");}
            UI::ShowNotification(Meta::ExecutingPlugin().Name, "No map for this number");
            status = Status::ManualReroll;
            return;
        }
        if (currentMap.url == "") {
            print("invalid map");
            UI::ShowNotification(Meta::ExecutingPlugin().Name, "Couldn't load map");
            status = Status::ManualReroll;
            return;
        }
        while (Time::Now < Wheel::targetTime) {
            sleep(Wheel::targetTime - Time::Now);
        }
        // load map (unless it was changed during roll, or is already open)
        if (currentMap !is null && currentMap.source is target) {
            CGameCtnChallenge@ rootMap = GetApp().RootMap;
            if (rootMap is null || currentMap.uid != rootMap.IdName) {
                openCurrentMap();
            }
            if (status == Status::RollingMap) {
                status = Status::Playing;
            }
        }
    }

    void openCurrentMap() {
        if (currentMap !is null) {
            startnew(CoroutineFunc(currentMap.play));
        } else {
            return;
        }
        status = Status::Playing;
    }

    string getMedalIcon() {
        return getMedalIcon(targetMedal);
    }
    string getMedalIcon(Medals medal) {
        if (medal == Medals::NoMedal) {
            return '\\$fff' + Icons::FlagCheckered;
        } else if(medal == Medals::Bronze) {
            return '\\$964' + Icons::Circle;
        } else if(medal == Medals::Silver) {
            return '\\$899' + Icons::Circle;
        } else if(medal == Medals::Gold) {
            return '\\$db4' + Icons::Circle;
        } else if(medal == Medals::Author) {
            return '\\$071' + Icons::Circle;
        } else if(medal == Medals::AutoBronze) {
            return '\\$964' + Icons::Circle;
        } else if(medal == Medals::AutoSilver) {
            return '\\$899' + Icons::Circle;
        } else if(medal == Medals::AutoGold) {
            return '\\$db4' + Icons::Circle;
        } else if(medal == Medals::AutoAuthor) {
            return '\\$071' + Icons::Circle;
        } else if(medal == Medals::WorldRecord) {
            return '\\$4bdWR';
        }
        return '?';
    }
    string getRerollMedalIcon() {
        if (targetMedal == Medals::NoMedal) {
            return '\\$fff Free';
        } else if(targetMedal == Medals::Bronze) {
            return getMedalIcon(Medals::NoMedal);
        } else if(targetMedal == Medals::Silver) {
            return getMedalIcon(Medals::Bronze);
        } else if(targetMedal == Medals::Gold) {
            return getMedalIcon(Medals::Silver);
        } else if(targetMedal == Medals::Author) {
            return getMedalIcon(Medals::Gold);
        } else if(targetMedal == Medals::AutoBronze) {
            return getMedalIcon(Medals::NoMedal);
        } else if(targetMedal == Medals::AutoSilver) {
            return getMedalIcon(Medals::AutoBronze);
        } else if(targetMedal == Medals::AutoGold) {
            return getMedalIcon(Medals::AutoSilver);
        } else if(targetMedal == Medals::AutoAuthor) {
            return getMedalIcon(Medals::AutoGold);
        } else if(targetMedal == Medals::WorldRecord) {
            return getMedalIcon(Medals::Author);
        }
        return '?';
    }
    string getMedalName() {
        return getMedalName(targetMedal);
    }
    string getMedalName(Medals medal) {
        if (medal == Medals::NoMedal) {
            return '\\$fffFinish';
        } else if(medal == Medals::Bronze) {
            return '\\$964Bronze';
        } else if(medal == Medals::Silver) {
            return '\\$899Silver';
        } else if(medal == Medals::Gold) {
            return '\\$db4Gold';
        } else if(medal == Medals::Author) {
            return '\\$071Author';
        } else if(medal == Medals::AutoBronze) {
            return '\\$964Auto Bronze';
        } else if(medal == Medals::AutoSilver) {
            return '\\$899Auto Silver';
        } else if(medal == Medals::AutoGold) {
            return '\\$db4Auto Gold';
        } else if(medal == Medals::AutoAuthor) {
            return '\\$071Auto Author';
        } else if(medal == Medals::WorldRecord) {
            return '\\$4bdWorld Record';
        }
        return 'Unknown medal';
    }
    string getRerollMedalName() {
        return getRerollMedalName(targetMedal);
    }
    string getRerollMedalName(Medals medal) {
        if (medal == Medals::NoMedal) {
            return '\\$fffFree';
        } else if(medal == Medals::Bronze) {
            return getMedalName(Medals::NoMedal);
        } else if(medal == Medals::Silver) {
            return getMedalName(Medals::Bronze);
        } else if(medal == Medals::Gold) {
            return getMedalName(Medals::Silver);
        } else if(medal == Medals::Author) {
            return getMedalName(Medals::Gold);
        } else if(medal == Medals::AutoBronze) {
            return getMedalName(Medals::NoMedal);
        } else if(medal == Medals::AutoSilver) {
            return getMedalName(Medals::AutoBronze);
        } else if(medal == Medals::AutoGold) {
            return getMedalName(Medals::AutoSilver);
        } else if(medal == Medals::AutoAuthor) {
            return getMedalName(Medals::AutoGold);
        } else if(medal == Medals::WorldRecord) {
            return getMedalName(Medals::Author);
        }
        return 'Unknown medal';
    }
}